#! /usr/bin/perl -W


# Copyright (C) 2000,2001 Mandrakesoft
# Author Sylvain de Tilly
#
# $Id: clean_rules.pl,v 1.4 2001/09/21 12:40:57 sdetilly Exp $
#
# This script cross through a rule list and remove rule due to
# options passed to it.



use strict;

use vars qw($opt_server $opt_client);
use Getopt::Long;

use lib qw(/usr/lib/naat/);
use backend::util;


my $Usage =<<"End_of_Usage;";
Clean all rules concerned by options :

Usage: clear_rules.pl transaction [options] list_rule
    - transaction : The transaction name
    - list_rule : The list you want to clean.

Just one of this options could be done but at least one.
    -s|--server : The server zone concerned.
    -c|--client : The client zone concerned.
End_of_Usage;


my $NULL_ITEM = "---\n";


# First get the transaction name (FIRST STEP IN ALL SCRIPTS)
my $transaction_name = shift @ARGV if (@ARGV && $ARGV[0] !~ /^-/);


if (! $transaction_name ) {
  $transaction_name = "$$";
  # initialize a backend transaction
  system ("config-wrapper.pl $transaction_name -i");
}


my $result = GetOptions ( 'server=s', 'client=s' );

if ( !$result ) { print STDERR $Usage; exit( 1 ); }

# list value to search in
my $list_value = $ARGV[ 0 ] if $ARGV[ 0 ];
if ( ! $list_value ) { print STDERR "No list defined...\n". $Usage; }


# The rule hash table.
my %rules_hash = ();


# print STDERR "args :\n\ttransname => $transaction_name\n\tclient => $opt_client\n\tserver => $opt_server\n";


# keys to clean :
my $keys;


my $RULES=`config-wrapper.pl $transaction_name --get $list_value`;
%rules_hash = backend::util::build_hash_from_list( $RULES );


if( ! ($opt_client || $opt_server) ) {
  print STDERR "You must passe one of source or client option...\n";
  print STDERR $Usage;
  exit( 1 );
}


if( $opt_client || $opt_server ) {
  foreach ( keys %rules_hash ) {
    if( $_ ne $NULL_ITEM ) {
      if ( ( $opt_client && $rules_hash{ $_ }{ "client" } eq $opt_client ) ||
	   ( $opt_server && $rules_hash{ $_ }{ "server" } eq $opt_server ) ) {
	$keys .= "$_,"
      }
    }
  }
}

if ( $keys ) {
  chop( $keys ); # remove the last ','
  `config-wrapper.pl $transaction_name --listdelkeys $list_value $keys`;
}
else {
  print STDERR "No modification to do in the list $list_value !\n";
}

# EOF
